#include "stdafx.h"
#include "input_tak.h"

////////////////////////////////////////////////////////////////////
// Component version information

// Forward declaration
static pfc::string_formatter g_get_component_about();

DECLARE_COMPONENT_VERSION_COPY(
	"TAK Decoder",
	"0.4.4",
	g_get_component_about()
	);

////////////////////////////////////////////////////////////////////
// Supported file type
// Used for "Open File" dialog and file associations.

DECLARE_FILE_TYPE("TAK Files", "*.TAK");

////////////////////////////////////////////////////////////////////
// Helper classes

/**
 * Helper class that formats the TAK library version.
 */
class format_tak_library_version
{
	pfc::string_formatter data;
public:
	explicit format_tak_library_version(TtakInt32 p_version)
	{
		data << pfc::format_int((p_version >> 16) & 0xff)
			<< "." << pfc::format_int((p_version >> 8) & 0xff)
			<< "." << pfc::format_int(p_version& 0xff);
	}

	operator char const *() {return data;}
};

/**
 * Generates the "About" text during component initialization.
 * The About text includes version information about the dynamically linked
 * TAK decoder library.
 */
static pfc::string_formatter g_get_component_about()
{
	pfc::string_formatter about;
	about << "Decodes and tags TAK files.";

	TtakInt32 version = 0, compatible_version = 0;
	if (tak_library::get_library_version(version, compatible_version) != tak_res_Ok)
	{
		about << "\n\nCould not retrieve version of TAK decoder library.";
	}
	else
	{
		about << "\n\nBuilt for TAK library version " << format_tak_library_version(tak_InterfaceVersion);
		about << "\nUsing TAK library version " << format_tak_library_version(version)
			<< " (compatible with versions down to " << format_tak_library_version(compatible_version) << ")";
	}

	about << "\n\nCopyright (c) 2007-2010 Holger Stenger\nTAK icon by Florian Trendelenburg (used with permission)";

	return about;
}

// Support for embedded cue sheets is provided by the chosen input factory.
static input_cuesheet_factory_t<input_tak> g_input_tak_factory;

////////////////////////////////////////////////////////////////////
// Embdded album art support (new in foobar2000 0.9.5)
// TAK uses APEv2 tags, so we can re-use the standard routines from
// the foobar2000 SDK.

class album_art_extractor_impl_tak : public album_art_extractor_impl_stdtags
{
public:
	album_art_extractor_impl_tak() : album_art_extractor_impl_stdtags("tak")
	{
	}
};

static service_factory_single_t<album_art_extractor_impl_tak> g_album_art_extractor_factory;
