#include "stdafx.h"
#include "input_tak.h"

#ifndef NDEBUG
#define TRACE(text) if (true) {console::formatter() << (text);} else {}
#else
#define TRACE(text)
#endif

////////////////////////////////////////////////////////////////////
// Settings for "Advanced" preferences page

// {C39E914E-05FB-4743-9FE5-AC50F10A2CE9}
static const GUID guid_branch_decoding_tak = { 0xc39e914e, 0x5fb, 0x4743, { 0x9f, 0xe5, 0xac, 0x50, 0xf1, 0xa, 0x2c, 0xe9 } };

// {82C21989-B97B-40ba-95A1-FA3269FB4AB8}
static const GUID guid_radio_tak_none = { 0x82c21989, 0xb97b, 0x40ba, { 0x95, 0xa1, 0xfa, 0x32, 0x69, 0xfb, 0x4a, 0xb8 } };

// {787EFCB4-EDAC-48b2-BF54-2D3D8159041D}
static const GUID guid_radio_tak_any = { 0x787efcb4, 0xedac, 0x48b2, { 0xbf, 0x54, 0x2d, 0x3d, 0x81, 0x59, 0x4, 0x1d } };

// {A6770E91-867A-4c49-92B7-CB4605B1FEDF}
static const GUID guid_radio_tak_asm = { 0xa6770e91, 0x867a, 0x4c49, { 0x92, 0xb7, 0xcb, 0x46, 0x5, 0xb1, 0xfe, 0xdf } };

// {85987D2C-7A54-47fd-BDAA-90D20167AD1E}
static const GUID guid_radio_tak_mmx = { 0x85987d2c, 0x7a54, 0x47fd, { 0xbd, 0xaa, 0x90, 0xd2, 0x1, 0x67, 0xad, 0x1e } };

// {5EE88C9C-758B-42f4-82ED-6BC255B6084D}
static const GUID guid_radio_tak_sse = { 0x5ee88c9c, 0x758b, 0x42f4, { 0x82, 0xed, 0x6b, 0xc2, 0x55, 0xb6, 0x8, 0x4d } };

static advconfig_branch_factory g_branch_decoding_tak("TAK decoder allowed optimizations", guid_branch_decoding_tak, advconfig_branch::guid_branch_decoding, 0.0);

static advconfig_radio_factory g_radio_tak_any("Any (default)", guid_radio_tak_any, guid_branch_decoding_tak, 1.0, true);
static advconfig_radio_factory g_radio_tak_none("None", guid_radio_tak_none, guid_branch_decoding_tak, 2.0, false);
static advconfig_radio_factory g_radio_tak_asm("ASM only", guid_radio_tak_asm, guid_branch_decoding_tak, 3.0, false);
static advconfig_radio_factory g_radio_tak_mmx("MMX only", guid_radio_tak_mmx, guid_branch_decoding_tak, 4.0, false);
static advconfig_radio_factory g_radio_tak_sse("SSE only", guid_radio_tak_sse, guid_branch_decoding_tak, 5.0, false);

////////////////////////////////////////////////////////////////////
// Helper classes

/**
 * Computes TAK CPU flags from advanced settings.
 */
static TtakInt32 g_get_cpu_opt_flags()
{
	TtakInt32 flags = tak_Cpu_Any;

	if (g_radio_tak_any.get_static_instance().get_state() == true)
		flags = tak_Cpu_Any;
	else if (g_radio_tak_none.get_static_instance().get_state() == true)
		flags = tak_Cpu_None;
	else if (g_radio_tak_asm.get_static_instance().get_state() == true)
		flags = tak_Cpu_Asm;
	else if (g_radio_tak_mmx.get_static_instance().get_state() == true)
		flags = tak_Cpu_MMX;
	else if (g_radio_tak_sse.get_static_instance().get_state() == true)
		flags = tak_Cpu_SSE;

	return flags;
}



void input_tak::open(service_ptr_t<file> p_filehint, const char * p_path, t_input_open_reason p_reason, abort_callback & p_abort)
{
	TRACE("input_tak::open()");
#ifndef NDEBUG
	m_run_called = false;
#endif

	m_file = p_filehint;
	// p_filehint may be null; this helper opens the file with appropriate privileges if necessary
	input_open_file_helper(m_file, p_path, p_reason, p_abort);

	//file_cached::g_create(m_file, m_file, p_abort, 1 << 20);

	m_decoder = new tak_lib_cpp::tak_seekable_decoder(m_file, g_get_cpu_opt_flags());
}

void input_tak::get_info(file_info & p_info, abort_callback & p_abort)
{
	TRACE("input_tak::get_info()");

	try
	{
		restore_file_position_helper helper(m_file, p_abort);

		// Read APEv2 tag.
		static_api_ptr_t<tag_processor_trailing>()->read(m_file, p_info, p_abort);
	}
	catch (exception_service_not_found)
	{
		// tag_processor_trailing service not present.
		console::formatter() << "APEv2 tag processor not present. Please make sure you have foo_input_std installed.";
	}
	catch (exception_tag_not_found)
	{
		// No APEv2 tag present. We can ignore this.
	}

	// Mark as lossless encoding.
	p_info.info_set("encoding", "lossless");
	
	m_decoder->get_info(p_info);

	if (p_info.get_length() > 0)
	try
	{
		t_filestats filestats = get_file_stats(p_abort);
		p_info.info_set_bitrate(t_int64((filestats.m_size * 8 / p_info.get_length() + 500) / 1000));
	}
	catch (exception_io)
	{
		// ignore
	}
}

t_filestats input_tak::get_file_stats(abort_callback & p_abort)
{
	TRACE("input_tak::get_file_stats()");

	return m_file->get_stats(p_abort);
}

void input_tak::decode_initialize(unsigned p_flags,abort_callback & p_abort)
{
	TRACE("input_tak::decode_initialize()");

	m_decoder->seek(0);

	m_sample_rate = m_decoder->get_sample_rate();
	m_bits_per_sample = m_decoder->get_bits_per_sample();
	m_channel_count = m_decoder->get_channel_count();

	const t_size bytes_per_block = ((m_bits_per_sample + 7) / 8) * m_channel_count;
	m_buffer.set_size(bytes_per_block * BUFFER_SIZE_IN_SAMPLES);
}

bool input_tak::decode_run(audio_chunk & p_chunk, abort_callback & p_abort)
{
#ifndef NDEBUG
	if (!m_run_called)
	{
		TRACE("input_tak::decode_run()");
		m_run_called = true;
	}
#endif

	const t_size bytes_per_block = ((m_bits_per_sample + 7) / 8) * m_channel_count;
	
	t_size sample_count = m_decoder->decode(m_buffer.get_ptr(), BUFFER_SIZE_IN_SAMPLES);

	if (sample_count == 0)
		return false;

	p_chunk.set_data_fixedpoint(
		m_buffer.get_ptr(),
		bytes_per_block * sample_count,
		m_sample_rate,
		m_channel_count,
		m_bits_per_sample,
		audio_chunk::g_guess_channel_config(m_channel_count));

	return true;
}

bool input_tak::decode_run_raw(audio_chunk & p_chunk, mem_block_container & p_raw, abort_callback & p_abort)
{
	TRACE("input_tak::decode_run_raw()");

	// TODO Implement method.
	throw pfc::exception_not_implemented();
}

void input_tak::decode_seek(double p_seconds, abort_callback & p_abort)
{
	TRACE("input_tak::decode_seek()");

	m_file->ensure_seekable();//throw exceptions if someone called decode_seek() despite of our input having reported itself as nonseekable.

	t_uint64 sample = audio_math::time_to_samples(p_seconds, m_decoder->get_sample_rate());

#ifdef TRACE_SEEKING
	console::formatter() << "decode_seek(" << p_seconds << ")";
#endif

	m_decoder->seek(sample);
}

bool input_tak::decode_can_seek()
{
	TRACE("input_tak::decode_can_seek()");

	// Check whether the underlying file is seekable.
	return m_file->can_seek();
}

bool input_tak::decode_get_dynamic_info(file_info & p_out, double & p_timestamp_delta)
{
	return false;
}

bool input_tak::decode_get_dynamic_info_track(file_info & p_out, double & p_timestamp_delta)
{
	return false;
}

void input_tak::decode_on_idle(abort_callback & p_abort)
{
	m_file->on_idle(p_abort);
}

void input_tak::retag(const file_info & p_info, abort_callback & p_abort)
{
	TRACE("input_tak::retag()");

	try
	{
		reopen_tak_decoder_helper(m_decoder, m_file, p_abort);

		// Write APEv2 tag.
		static_api_ptr_t<tag_processor_trailing>()->write_apev2(m_file, p_info, p_abort);
	}
	catch (exception_service_not_found)
	{
		console::formatter() << "APEv2 tag processor not present. Please make sure you have foo_input_std installed.";
	}
}

bool input_tak::g_is_our_content_type(const char * p_content_type)
{
	return false;
}

bool input_tak::g_is_our_path(const char * p_path,const char * p_extension)
{
	return stricmp_utf8(p_extension, "tak") == 0;
}

void input_tak::set_logger(event_logger::ptr ptr)
{
	TRACE("input_tak::set_logger()");

	// TODO Implement method.
	m_logger = ptr;
	m_decoder->set_logger(ptr);
}
